package com.gic.download.utils;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.serializer.SerializerFeature;
import com.gic.download.constants.ExcelExtensionEnum;
import com.gic.download.dto.DownloadReportTempDTO;
import com.gic.download.qo.DownloadExcelQO;
import com.gic.download.qo.HeaderQO;
import com.gic.thirdparty.BucketNameEnum;
import com.gic.thirdparty.FileOperateUtils;
import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.beanutils.PropertyUtilsBean;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.client.utils.DateUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.ss.usermodel.*;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

import javax.servlet.http.HttpServletResponse;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.text.SimpleDateFormat;
import java.util.*;
/**
 * 下载工具
 * @ClassName: DownloadUtils 
 * @Description:  
 * @author guojuxing 
 * @date 2020/9/15 10:43 AM 
 */
public class DownloadUtils {
    private static final Logger LOGGER = LogManager.getLogger(DownloadUtils.class);

    private static final String SUFFIX_EXCEL_2007 = "xlsx";
    private static final String SUFFIX_CSV = "csv";
    private static final String SUFFIX_EXCEL_2003 = "xls";


    /**
     * 分页下载数据获取类
     * @author Silver
     * @date 2017年3月16日 上午11:45:13
     *
     * @param <T>
     *
     */
    protected abstract class DownloadDataLoader<T> {

        /**
         * 分页下载属性值控制操作类
         * @param bean
         * @param propertyName
         * @param property
         * @return
         * @author Silver
         * @date 2017年3月16日 上午11:45:45
         */
        protected String convertProperty(T bean, String propertyName, Object property) {
            if (property == null) {
                return "";
            }
            try {
                Class clazz = getPropertyType(bean, propertyName);
                if (Date.class.equals(clazz)) {
                    //如果是日期类型
                    SimpleDateFormat format = new SimpleDateFormat("E MMM dd hh:mm:ss z yyyy", Locale.US);

                    SimpleDateFormat resultFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
                    return resultFormat.format(format.parse(property.toString()));
                }
            } catch (Exception e) {
                LOGGER.info("bean:" + bean + ",Property:" + propertyName + e.getMessage(), e);
                return null;
            }
            return property.toString();
        }

        protected Class getPropertyType(T bean, String propertyName) {
            try {
                return new PropertyUtilsBean().getPropertyType(bean, propertyName);
            } catch (Throwable e) {
                LOGGER.info("bean:" + bean + ",Property:" + propertyName + e.getMessage(), e);
                return null;
            }
        }

        /**
         * 分页下载属性赋值
         * @param bean
         * @param propertyName
         * @return
         * @author Silver
         * @date 2017年3月16日 上午11:46:37
         */
        protected Object getProperty(T bean, String propertyName) {
            try {
                return BeanUtils.getProperty(bean, propertyName);
            } catch (Throwable e) {
                LOGGER.info("bean:" + bean + ",Property:" + propertyName + e.getMessage(), e);
                return null;
            }
        }

        /**
         * 用于二级表头
         * @param bean
         * @param propertyName
         * @return
         */
        protected Object getPropertyOfDouble(Object bean, String propertyName) {
            try {
                if (bean == null) {
                    return null;
                }
                if (bean instanceof Map) {
                    //如果是map
                    return ((Map) bean).get(propertyName);
                }
                return BeanUtils.getProperty(bean, propertyName);
            } catch (Throwable e) {
                LOGGER.info("bean:" + bean + ",Property:" + propertyName + e.getMessage(), e);
                return null;
            }
        }

        protected String convertPropertyOfDouble(Object bean, String propertyName, Object property) {
            if (property == null) {
                return "";
            }
            return property.toString();
        }

        /**
         * 数据获取接口
         * @param pageNum -- 从0计数
         * @return
         * @throws Exception
         * @author Silver
         * @date 2017年3月16日 上午11:47:07
         */
        protected abstract List<T> getDownloadData(Integer pageNum) throws Exception;
    };


    protected static interface Writer {
        /**
         * 写数据
         * @Title: write 
         * @Description:
         * @author guojuxing
         * @param row 
         * @return  void 
         * @throws IOException
         */
        public void write(Collection<String> row) throws IOException;
    }

    /**
     * Web下载文件
     * @param tempPath 临时路径
     * @param reportId 报表中心ID
     * @param fileName 文件名称
     * @param excelExtensionCode Excel文件扩展名 枚举 ExcelExtensionEnum
     * @param headerList 表头名称
     * @param propertyNameList 字段名
     * @param loader    数据加载类
     * @param needEncryptField 需要加密的字段，每一个元素存的是字段，如phone（电话）
     * @param columnWidth 列宽
     * @throws Exception
     * @author Silver
     * @date 2017年3月16日 上午11:47:31
     */
    protected <T> void download(String tempPath, Integer reportId, String fileName, Integer excelExtensionCode, List<String> headerList, List<String> propertyNameList, DownloadDataLoader<T> loader, List<String> needEncryptField, List<Integer> columnWidth) throws Exception {
        downloadCommon(new DownloadExcelQO().setTempPath(tempPath).setReportId(reportId).setFileName(fileName)
                .setExcelExtensionCode(excelExtensionCode).setHeaderList(headerList).setPropertyNameList(propertyNameList)
                .setNeedEncryptField(needEncryptField).setColumnWidth(columnWidth), loader);
    }

    /**
     * 下载文件，2层标题结构
     * @Title: downloadOfDoubleHeaderTitle 
     * @Description:
     * @author guojuxing
     * @param tempPath 临时路径
     * @param reportId 报表中心ID
     * @param fileName 文件名称
     * @param excelExtensionCode Excel文件扩展名 枚举 ExcelExtensionEnum
     * @param headerList 表头名称 2层标题结构
     * @param propertyNameList 字段名
     * @param loader 数据加载类
     * @param needEncryptField 需要加密的字段，每一个元素存的是字段，如phone（电话）
     * @param columnWidth 列宽
     * @return  void 
     */
    @Deprecated
    protected <T> void downloadOfDoubleHeaderTitle(String tempPath, Integer reportId, String fileName, Integer excelExtensionCode, List<HeaderQO> headerList, List<String> propertyNameList, DownloadDataLoader<T> loader, List<String> needEncryptField, List<Integer> columnWidth) throws Exception {
        downloadCommon(new DownloadExcelQO().setTempPath(tempPath).setReportId(reportId).setFileName(fileName)
                .setExcelExtensionCode(excelExtensionCode).setHeaders(headerList).setPropertyNameList(propertyNameList)
                .setNeedEncryptField(needEncryptField).setColumnWidth(columnWidth), loader);
    }

    /**
     * 下载文件，2层标题结构
     * @Title: downloadForDoubleHeaderTitle 
     * @Description:
     * @author guojuxing
     * @param tempPath 临时路径
     * @param reportId 报表中心ID
     * @param fileName 文件名称
     * @param excelExtensionCode Excel文件扩展名 枚举 ExcelExtensionEnum
     * @param headerList 表头名称 2层标题结构
     * @param propertyNameList 字段名
     * @param loader 数据加载类
     * @param needEncryptField 需要加密的字段，每一个元素存的是字段，如phone（电话）
     * @param columnWidth 列宽
     * @return  void 
     */
    protected <T> void downloadForDoubleHeaderTitle(String tempPath, Integer reportId, String fileName, Integer excelExtensionCode, LinkedHashMap<String, List<String>> headerList, LinkedHashMap<String, List<String>> propertyNameList, DownloadDataLoader<T> loader, Map<String, List<String>> needEncryptField, List<Integer> columnWidth) throws Exception {
        downloadCommon(new DownloadExcelQO().setTempPath(tempPath).setReportId(reportId).setFileName(fileName)
                .setExcelExtensionCode(excelExtensionCode).setHeaderMap(headerList).setPropertyNameMap(propertyNameList)
                .setNeedEncryptFieldMap(needEncryptField).setColumnWidth(columnWidth), loader);
    }

    private <T> void downloadCommon(DownloadExcelQO param, DownloadDataLoader<T> loader) throws Exception {
        LOGGER.info("下载导出数据参数:{}", JSON.toJSONString(param, SerializerFeature.WriteMapNullValue));

        Integer reportId = param.getReportId();
        String fileName = param.getFileName();
        Integer excelExtensionCode = param.getExcelExtensionCode();
        List<String> headerList = param.getHeaderList();
        List<String> propertyNameList = param.getPropertyNameList();
        List<String> needEncryptField = param.getNeedEncryptField();
        String tempPath = param.getTempPath();
        //是否是2层标题结构
        LinkedHashMap<String, List<String>> headerMap = param.getHeaderMap();
        LinkedHashMap<String, List<String>> propertyNameMap = param.getPropertyNameMap();
        Map<String, List<String>> needEncryptFieldMap = param.getNeedEncryptFieldMap();
        boolean isDoubleHeaderTitle = headerMap != null && !headerMap.isEmpty();

        if (StringUtils.isEmpty(fileName) || loader == null) {
            throw new RuntimeException("参数错误。FileName:" + fileName + ",DataLoader:" + loader);
        }
        // 获取输出流,设置content-type等头域
        //final OutputStream out = getResponseStream(response ,fileName);
        File tempFile = null;
        //重置临时文件路径
        DataDownloadUtils.mkDir(new File(tempPath));
        File file = new File(tempPath);
        //初始化数据
        int leastPrefixStringLength = 3;
        if (fileName.length() < leastPrefixStringLength) {
            //如果长度太短
            fileName = fileName + "__";
        }
        tempFile = File.createTempFile(fileName, ExcelExtensionEnum.getExtensionByCode(excelExtensionCode), file);
        final OutputStream out = new FileOutputStream(tempFile);
        try {
            Writer writer = null;
            // 获取文件后缀名
            String extension = ExcelExtensionEnum.getExtensionByCode(excelExtensionCode);
            // 如果是excel的后缀
            if (ExcelExtensionEnum.EXCEL_2003.getExtension().equalsIgnoreCase(extension) || ExcelExtensionEnum.EXCEL_2007.getExtension().equalsIgnoreCase(extension)) {
                Workbook workbook;
                if (ExcelExtensionEnum.EXCEL_2003.getExtension().equalsIgnoreCase(extension)) {
                    workbook = new HSSFWorkbook();
                } else {
                    workbook = new XSSFWorkbook();
                }
                Sheet sheet = workbook.createSheet("sheet1");
                final List<Collection<String>> rows = new ArrayList<>();
                writer = new Writer() {
                    @Override
                    public void write(Collection<String> row) {
                        rows.add(row);
                    }
                };

                if (isDoubleHeaderTitle) {
                    //二级表头处理
                    writeOutputStreamOfDoubleHeader(loader, propertyNameMap, writer, needEncryptFieldMap, reportId);
                } else {
                    writeOutputStream(loader, propertyNameList, writer, needEncryptField, reportId);
                }

                // 写入excel
                if (isDoubleHeaderTitle) {
                    CellStyle style = workbook.createCellStyle();
                    //// 指定单元格居中对齐
                    style.setAlignment(HorizontalAlignment.CENTER);
                    // 指定单元格垂直居中对齐
                    style.setVerticalAlignment(VerticalAlignment.CENTER);
                    if (!ExcelUtils.doubleHeaderTitle(style, sheet, param.getColumnWidth(), headerMap, rows)) {
                        throw new IOException("设置导出文件内容失败。");
                    }
                } else {
                    if (!ExcelUtils.setExcelInfo(sheet, param.getColumnWidth(), headerList, rows)) {
                        throw new IOException("设置导出文件内容失败。");
                    }
                }

                workbook.write(out);
            } else if (ExcelExtensionEnum.CSV.getExtension().equalsIgnoreCase(extension)) {
                writer = new Writer() {
                    @Override
                    public void write(Collection<String> row) throws IOException {
                        String str = ExcelUtils.collectionToCsvString(row);
                        byte[] content = org.apache.commons.codec.binary.StringUtils.getBytesUnchecked(str + "\n",
                                "GBK");
                        IOUtils.write(content, out);
                        out.flush();
                    }
                };
                // 写文件头
                writer.write(headerList);
                // 写文件
                writeOutputStream(loader, propertyNameList, writer, needEncryptField, reportId);
            } else {
                writer = new Writer() {
                    @Override
                    public void write(Collection<String> row) throws IOException {
                        IOUtils.write(org.apache.commons.codec.binary.StringUtils.getBytesUnchecked(row + "\n",
                                "GBK"), out);
                        out.flush();
                    }
                };
                // 写文件头
                writer.write(headerList);
                // 写文件
                writeOutputStream(loader, propertyNameList, writer, needEncryptField, reportId);
            }
            out.flush();
            tempFile.deleteOnExit();
            //上传文件
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss");
            String fieldCode = "/" + fileName + "_" + sdf.format(new Date()) + ExcelExtensionEnum.getExtensionByCode(excelExtensionCode);
            LOGGER.info("上传文件到腾讯云，文件名：" + fieldCode);
            //文件存进腾讯云
            String url = FileOperateUtils.simpleUploadFileFromLocal(tempFile, fieldCode, BucketNameEnum.REPORT_50000.getName());
            if (org.apache.commons.lang.StringUtils.isNotBlank(url)) {
                url = "https://" + url;
            }
            LOGGER.info("上传文件到腾讯云，路径：" + url);
            //更新数据
            DataDownloadUtils.uploadFile(reportId, url);
        } finally {
            IOUtils.closeQuietly(out);
        }
    }

    /**
     * 获得输出流
     *
     * @return
     * @throws IOException
     */
    protected OutputStream getResponseStream(HttpServletResponse response, String fileName) throws IOException {
        if (StringUtils.isEmpty(fileName)) {
            fileName = DateUtils.formatDate(new Date(), "yyyy-MM-dd_HH-mm-ss") + SUFFIX_CSV;
        }
        response.reset();

        String extension = FilenameUtils.getExtension(fileName);
        if(SUFFIX_EXCEL_2007.equalsIgnoreCase(extension)){
            // 部分window版本生成后的xlsx打不开，默认改成xls打开
            fileName = fileName.substring(0, fileName.length() -1);
        }
        //设置响应编码
        response.setCharacterEncoding("UTF-8");
        //设置对应的contentType
        response.setContentType("application/x-download;charset=UTF-8");
        response.setHeader("Content-Disposition", "attachment; filename="
                + new String(fileName.getBytes("gb2312"), "ISO-8859-1"));
        OutputStream out = response.getOutputStream();


        return out;
    }

    private void test() {
        LinkedHashMap<String, List<String>> headerMap = new LinkedHashMap<>(16);
        headerMap.put("测试1", null);
        headerMap.put("域名2", Arrays.asList("门店1", "门店2"));
        headerMap.put("域名3", Arrays.asList("门店1", "门店2"));
        LinkedHashMap<String, List<String>> headerPropertyNameMap = new LinkedHashMap<>(16);
        headerPropertyNameMap.put("test1", null);
        headerPropertyNameMap.put("2", Arrays.asList("store1", "store2"));
        headerPropertyNameMap.put("3", Arrays.asList("store1", "store2"));
        Map<String, List<String>> needEncryptFieldMap = new HashMap<>(16);
        needEncryptFieldMap.put("test1", null);
        needEncryptFieldMap.put("2", Arrays.asList("store1"));
        needEncryptFieldMap.put("3", Arrays.asList("store1"));
        List<Map<String, Object>> dataList = new ArrayList<>();

    }

    protected <T> void writeOutputStream(DownloadDataLoader<T> loader, List<String> propertyNames,
                                         Writer writer, List<String> needEncryptField, Integer reportId) throws Exception {

        DownloadReportTempDTO tempReport = DataDownloadUtils.getByReportId(reportId);
        int pageNum = 1;
        int maxLength = 102400;
        while (maxLength -- > 0) {
            // 分页获取数据
            List<T> objList = null;
            try {
                objList = loader.getDownloadData(pageNum++);
            } catch (Exception e) {
                LOGGER.error("获得到处数据异常:{}",e.getMessage(), e);
            }
            if (CollectionUtils.isEmpty(objList)) {
                break;
            }
            for (T bean : objList) {
                if (bean == null) {
                    continue;
                }
                Collection<String> result = new ArrayList<String>();
                // 遍历指定属性
                for (String name : propertyNames) {
                    // 获得属性值
                    Object property = loader.getProperty(bean, name);
                    // 将属性值转换成字符串
                    String convertValue = loader.convertProperty(bean, name, property);
                    //加密处理
                    //如果不是完整数据，则需要加密处理
                    if (tempReport != null && tempReport.getDataType().intValue() != 1) {
                        if (needEncryptField != null && needEncryptField.contains(name)) {
                            convertValue = encrypt(convertValue);
                        }
                    }
                    // 组装成row
                    result.add(convertValue);
                }
                if (CollectionUtils.isEmpty(result)) {
                    continue;
                }
                writer.write(result);
            }
        }

    }

    protected <T> void writeOutputStreamOfDoubleHeader(DownloadDataLoader<T> loader, LinkedHashMap<String, List<String>> propertyNames,
                                         Writer writer, Map<String, List<String>> needEncryptField, Integer reportId) throws Exception {

        DownloadReportTempDTO tempReport = DataDownloadUtils.getByReportId(reportId);
        //是否是否加密处理
        boolean isNeedEncrypt = tempReport != null && tempReport.getDataType().intValue() != 1;
        boolean hasNeedEncryptField = needEncryptField != null && !needEncryptField.isEmpty();
        int pageNum = 1;
        int maxLength = 102400;
        while (maxLength -- > 0) {
            // 分页获取数据
            List<T> objList = null;
            try {
                objList = loader.getDownloadData(pageNum++);
            } catch (Exception e) {
                LOGGER.error("获得到处数据异常:{}",e.getMessage(), e);
            }
            if (CollectionUtils.isEmpty(objList)) {
                break;
            }
            for (T bean : objList) {
                if (bean == null) {
                    continue;
                }
                //查询的数据对象，格式转换
                Map<String, Object> obj = (Map<String, Object>) bean;

                Collection<String> result = new ArrayList<String>();
                // 遍历指定属性
                for (Map.Entry<String, List<String>> entry : propertyNames.entrySet()) {
                    // 获得属性值 （表头第二层数据）
                    List<String> propertyValue = entry.getValue();
                    //数据对象的value
                    Object objValue = obj.get(entry.getKey());

                    if (CollectionUtils.isEmpty(propertyValue)) {
                        //如果是空，说明只有一层
                        String value = Optional.ofNullable(objValue).orElse("").toString();
                        if (isNeedEncrypt && hasNeedEncryptField) {
                            if (needEncryptField.containsKey(entry.getKey())) {
                                value = encrypt(value);
                            }
                        }
                        result.add(value);
                    } else {
                        //如果不为空，说明是二层结构，则
                        for (String pro : propertyValue) {
                            // 获得属性值
                            Object property = loader.getPropertyOfDouble(objValue, pro);
                            // 将属性值转换成字符串
                            String convertValue = loader.convertPropertyOfDouble(objValue, pro, property);
                            if (isNeedEncrypt && hasNeedEncryptField) {
                                boolean hasProperty = needEncryptField.containsKey(entry.getKey())
                                        && CollectionUtils.isNotEmpty(needEncryptField.get(entry.getKey()))
                                        && needEncryptField.get(entry.getKey()).contains(pro);
                                if (hasProperty) {
                                    convertValue = encrypt(convertValue);
                                }
                            }
                            result.add(convertValue);
                        }
                    }
                }
                if (CollectionUtils.isEmpty(result)) {
                    continue;
                }
                writer.write(result);
            }
        }

    }

    private static String encrypt(String data) {
        if (StringUtils.isBlank(data)) {
            return data;
        }
        //手机号或者会员卡号 第5-8位用*代替
        int length = data.length();

        boolean isGreaterThanEight = length > 8;
        boolean isGreaterThanFour = length > 4;
        if (isGreaterThanEight) {
            return data.substring(0, 4) + "****" + data.substring(8);
        } else if (isGreaterThanFour) {
            return data.substring(0, 4) + "****";
        } else {
            return "****";
        }
    }

}
